/*
 * This file is part of PB: PerfBench.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 *  Instrument C language programs with pb calls.
 *  Grammar based on "C, A Reference Manual", Harbison & Steele, 3rd edition.
 *  Implemented as two finite state machines: one to perform the lexical
 *  analysis (gettoken) and one to parse the input code and instrument at
 *  appropriate locations (fsm in main()).
 *
 * TODOs:
 * 1. check for a second arg to main, if it exists, use first element of
 *    array as name in PBinit,PBstart,PBend for main().
 * 2. if a return statement has an expression (e.g., a recursive call)
 *    should it be evaluated prior to the call to PBend?
 */

  /* 
    $RCSfile: cpb.c,v $
    $Revision: 1.6 $ 
  */ 

#include <stdio.h>
#include <string.h>
#include "cpb.h"

static char* progname = 0;
static void usage() {
  fprintf(stderr, "usage: %s [-f list] [file]\n", progname);
  exit(1);
}
static void prstart(const char* fdec);
static void prend(const char* fdec);
static int inlist(const char* fdec);
static void mklist(const char* file);

int main(int argc, char** argv) 
{
  FILE *fp = 0;
  char word[BUFSIZ];
  char fdec[BUFSIZ];
  int lb = 0; /* brace counter */
  int lp = 0; /* paren counter */
  int c;
  Token t;
  enum {NEUTRAL, IDENT, FDECL, CSTMT} state = NEUTRAL;
  extern int lineno;
  progname = argv[0];

  fprintf(stderr,"After generating PB instrumented files, please manually add\n"
	         "a single INITPB(fileName, EventStr) in main() or wherever it is appropriate\n"
	         "\n"
	         "Also add a call to REPORTPB after the last ENDPB call\n");

  switch (argc) 
    {
    case 1:
      fp = stdin;
      break;
    case 2:
      if ((fp = fopen(argv[1], "r")) == 0) {
	fprintf(stderr,"%s: can't open: %s\n",progname,argv[1]);
	exit(1);
      }
      break;
    case 3:
      if (strcmp(argv[1], "-f") != 0)
	usage();
      fp = stdin;
      mklist(argv[2]);
      break;
    case 4:
      if (strcmp(argv[1], "-f") != 0)
	usage();
      mklist(argv[2]);
      if ((fp = fopen(argv[3], "r")) == 0) 
	{
	  fprintf(stderr,"%s: can't open: %s\n",progname,argv[3]);
	  exit(1);
	}
      break;
    default:
      usage();
    }
  printf("/* To Use PB one must define USE_PB, It can be handled via a compiler option. */\n"
	 "#define USE_PB\n"
	 "#include \"pbUser.h\"\n");
  while ((t = gettoken(word, fp)) != TEOF) 
    {
      if (!(state == CSTMT && (t == TRBRACE || t == TRESERVED)))
	prtoken(t, word), fflush(stdout);
      switch (state) 
	{
	case NEUTRAL:
	  switch (t) 
	    {
	    case TIDENT:
	      state = IDENT;
	      strcpy(fdec, word);
	      continue;
	    case TRESERVED:
	      if (strcmp(word, "return") == 0) 
		{
		  fprintf(stderr, "'return' not understood\n");
		  if (fp != stdin)
		    fprintf(stderr, "file: %s ", argv[1]);
		  fprintf(stderr, "line: %d\n", lineno);
		  exit(1);
		}
	    }
	case IDENT:
	  if (t == TLPAREN) 
	    {
	      state = FDECL;
	      lp = 1;
	    } 
	  else 
	    {
	      state = NEUTRAL;
	    }
	  continue;
	case FDECL:
	  switch (t) 
	    {
	    case TLPAREN:
	      lp++;
	      continue;
	    case TRPAREN:
	      lp--;
	      if (lp == 0) 
		{
		  lb = 0;
		  state = CSTMT;
		}
	      continue;
	    }
	  continue;
	case CSTMT:
	  switch (t) 
	    {
	    case TLBRACE:
	      lb++;
	      if (lb == 1 && inlist(fdec))
		prstart(fdec);
	      continue;
	    case TRBRACE:
	      lb--;
	      if (lb == 0) 
		{
		  if (inlist(fdec))
		    prend(fdec);
		  state = NEUTRAL;
		}
	      prtoken(t, word);
	      continue;
	    case TRESERVED:
	      if (strcmp(word, "return") == 0 ||
		  strcmp(word, "exit") == 0) 
		{
		  int c;
		  if (inlist(fdec)) 
		    {
		      printf("do {");
		      prend(fdec);
		    }
		  prtoken(t, word);
		  while (c = fgetc(fp)) 
		    {
		      if (c == EOF) 
			{
			  fprintf(stderr,
				  "unexpected eof\n");
			  exit(1);
			}
		      printf("%c", c);
		      if (c == ';') 
			{
			  if (inlist(fdec))
			    printf("} while (0);\n");
			  break;
			}
		    }
		  continue;
		}
	      prtoken(t, word);
	      continue;
	    }
	}
    }
}

static void prstart(const char* fdec) 
{
  if (strcmp(fdec, "main") == 0) {
    printf("/* Put something like this: \n"
	   "    INITPB(outputFileName,\"flops\"); \n"
           " in this routine  before the STARTPB call */\n" );
  }
  /* emit pbstart with fdecl */
  printf("STARTPB(\"%s\");\n", fdec);
  return;
}

static void prend(const char* fdec) 
{
  printf("ENDPB(\"%s\");\n", fdec);
  return;
}

static struct flist 
{
  char *fname;
  struct flist *next;
} *flist = 0;

static void mklist(const char* file) 
{
  FILE* fp = 0;
  char buf[BUFSIZ];

  if ((fp = fopen(file, "r")) == 0) 
    {
      fprintf(stderr, "can't open function list file: %s\n", file);
      exit(1);
    }
  while (fgets(buf, sizeof(buf), fp)) 
    {
      struct flist *f = (struct flist*)malloc(sizeof(struct flist));
      if (f == 0)
	perror(progname), exit(1);
      buf[strlen(buf)-1] = '\0';
      if ((f->fname = (char*)malloc(sizeof(char)*(strlen(buf)+1)))==0)
	perror(progname), exit(1);
      strcpy(f->fname, buf);
      f->next = flist;
      flist = f;
    }
  return;
}

static int inlist(const char* fdec) 
{
  struct flist *p = 0;
  if (strcmp(fdec, "main") == 0)
    return 1;
  if (flist == 0)
    return 1;
  for (p = flist; p; p = p->next)
    if (strcmp(fdec, p->fname) == 0)
      return 1;
  return 0;
}
