/*
 * 
 * Copyright (c) 2001-2002 Robert McLay
 *
 * Authors: Robert McLay <mclay@cfdlab.ae.utexas.edu>
 *          Victor Parr  <vparr@mindspring.com>
 */

/*
 * This file is part of PB: PerfBench.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include "pb.h"
#include "hash.h"
#define VN "hash.c : $Revision: 1.7 $ $Date: 2002/05/22 02:24:41 $\n"

Hash* make_hashtable(int size, 
                     unsigned int (*hash_fn)(const char *),
                     int (*match_fn)(const char *, const char *),
                     size_t (*length_fn)(const char *))
{
  Hash* ht;                     /* working hash table context (object) */

  ALLOC(ht, sizeof(Hash), Hash*);
  ht->num_buckets = size;
  ht->hash_fn     = hash_fn;
  ht->match_fn    = match_fn;
  ht->length_fn   = length_fn;
  ht->num_entries = 0;

  {
    Buckets *buckets;             /* bucket storage */
    int     ja;                   /* counter */
    ALLOC(buckets, sizeof(Buckets)*size, Buckets *);
    for (ja = 0; ja < size; ja++)
      {
        buckets[ja].num_entries = 0;
        buckets[ja].link        = NULL;
      }
    ht->buckets = buckets;
  }

  return ht;
}

void free_hashtable(Hash* ht, int flag)
{
  {
    int   ja;
    Link* head;
    Link* link;
    for (ja = 0; ja < ht->num_buckets; ja++)
      for (head = ht->buckets[ja].link; head; )
        {
          link = head;
          head = link->next;
          free(link->key);
          if (flag) free(link->adt);
          free(link);
        }
  }

  free(ht->buckets);  ht->buckets = NULL;
  free(ht);           ht          = NULL;
}

void* insert_in_hash(Hash* ht, char* key, void* adt, void* result)
{
  int   location = (*(ht->hash_fn))(key) % ht->num_buckets;
  Link* link;

  for (link=ht->buckets[location].link; link; link = link->next)
    if ( ! (*(ht->match_fn))(key,link->key) )
      {
        result = link->adt;
        return NULL;
      }


  {
    Link* next = ht->buckets[location].link;
    int   len;
    ALLOC(link, sizeof(Link), Link *); /* Create link */
    len = (*(ht->length_fn))(key);     /* copy key into [[link]] */
    ALLOC(link->key, len+1, char *);
    memcpy(link->key, key, len+1);
    link->adt = adt;                   /* save [[adt]] into [[link]] */
    result    = adt;                   /* store [[adt]] into [[result]] */

    ht->buckets[location].link = link; /* save new link to head of list */
    link->next                 = next; /* and connect it to rest of list */
    ht->buckets[location].num_entries++;
    ht->num_entries++;

    return adt;
  }

}

void* find_in_hash(Hash* ht, char* key)
{
  Link* link;
  int   location = (*(ht->hash_fn))(key) % ht->num_buckets;

  for (link=ht->buckets[location].link; link; link = link->next)
    if ( ! (*(ht->match_fn))(key,link->key) )
      return link->adt;
  return NULL;
}

void for_all_in_hash(Hash* ht, void (*func)(const char *, void *))
{
  int   ja;
  Link* link;
  for (ja = 0; ja < ht->num_buckets; ja++)
    for (link = ht->buckets[ja].link; link; link=link->next )
      (*func)(link->key, link->adt);
}

#ifndef HAVE_STRCASECMP
#define TOLOWER(c)  (( isupper(c) ) ? tolower(c) : (c))
int strcasecmp( const char *str1, const char *str2 )
{
   register  const char *s1 = str1, *s2 = str2;
   register  char  c1, c2;

   if ( s1 == s2 )  return   0;
   if ( !s1 )       return  -1;
   if ( !s2 )       return   1;

   for ( ; *s1 && *s2 ; s1++ , s2++ ) {
      c1 = TOLOWER( *s1 );
      c2 = TOLOWER( *s2 );

      if (c1 != c2)  return  (int)(c1 -c2);
   }
   return   (*s1 == *s2) ? 0 : (int)(*s1 - *s2);
}
#endif

#define BITS_IN_int    (sizeof (int)*8)
#define THREE_QUARTERS ((int) ((BITS_IN_int * 3) / 4))
#define ONE_EIGHTH     ((int) (BITS_IN_int / 8))
#define HIGH_BITS      ( ~ ((unsigned int) (~0) >> ONE_EIGHTH ))

unsigned int hashUC(const char * string)
{
  unsigned int hash_value, i;
  for ( hash_value = 0; *string; ++string)
    {
      hash_value = (hash_value << ONE_EIGHTH) + toupper(*string);
      if ( (i = hash_value & HIGH_BITS ) != 0)
        hash_value = (hash_value ^ (i >> THREE_QUARTERS )) & ~HIGH_BITS;
    }
  return hash_value;
}



